@@ -42,8 +42,15 @@ async def _build_oa_chat_request(
             # Disable reasoning for Gemini models when no effort requested
             if "gemini" in model.id:
                 effort = "none"
+            elif "gpt-5" in model.id:
+                effort = "minimal"
             else:
                 effort = "low"
+        if effort == "minimal" and "gpt-5" not in model.id:
+            print(
+                "WARNING: 'minimal' reasoning effort only allowed for gpt-5. setting to 'low'."
+            )
+            effort = "low"
         request_json["reasoning_effort"] = effort
     else:
         if sampling_params.reasoning_effort:
@@ -106,6 +113,9 @@ class OpenAIRequest(APIRequestBase):
         finish_reason = None
         assert self.context.status_tracker
 
+        if status_code == 500:
+            print("Internal Server Error: ", (await http_response.text()))
+
         if status_code >= 200 and status_code < 300:
             try:
                 data = await http_response.json()
@@ -122,15 +132,21 @@ class OpenAIRequest(APIRequestBase):
                     message = data["choices"][0]["message"]
                     finish_reason = data["choices"][0]["finish_reason"]
 
-                    # Add text content if present
-                    if message.get("content"):
-                        parts.append(Text(message["content"]))
-
                     # Add thinking content if present (reasoning models)
                     if "reasoning_content" in message:
                         thinking = message["reasoning_content"]
                         parts.append(Thinking(thinking))
 
+                    # Together AI returns reasoning in a "reasoning"
+                    # field which is not correct but whatever
+                    if message.get("reasoning"):
+                        thinking = message["reasoning"]
+                        parts.append(Thinking(thinking))
+
+                    # Add text content if present
+                    if message.get("content"):
+                        parts.append(Text(message["content"]))
+
                     # Add tool calls if present
                     if "tool_calls" in message:
                         for tool_call in message["tool_calls"]:
@@ -292,6 +308,9 @@ class OpenAIResponsesRequest(APIRequestBase):
         data = None
         assert self.context.status_tracker
 
+        if status_code == 500:
+            print("Internal Server Error: ", http_response.text())
+
         if status_code >= 200 and status_code < 300:
             try:
                 data = await http_response.json()
@@ -415,10 +434,12 @@ class OpenAIResponsesRequest(APIRequestBase):
                     error_message = f"Error parsing {self.model.name} responses API response: {str(e)}"
 
         elif mimetype and "json" in mimetype.lower():
+            print("is_error True, json response")
             is_error = True
             data = await http_response.json()
             error_message = json.dumps(data)
         else:
+            print("is_error True, non-json response")
             is_error = True
             text = await http_response.text()
             error_message = text
