@@ -1,10 +1,9 @@
 import asyncio
-import random
-from typing import Any, Literal, Self, Sequence, overload
+from typing import Any, AsyncGenerator, Callable, Literal, Self, Sequence, overload
 
 import numpy as np
 import yaml
-from pydantic import BaseModel
+from pydantic import BaseModel, PrivateAttr
 from pydantic.functional_validators import model_validator
 
 from lm_deluge.api_requests.openai import stream_chat
@@ -23,22 +22,14 @@ from .request_context import RequestContext
 from .tracker import StatusTracker
 
 
-# TODO: get completions as they finish, not all at once at the end.
 # TODO: add optional max_input_tokens to client so we can reject long prompts to prevent abuse
-class LLMClient(BaseModel):
+class _LLMClient(BaseModel):
     """
-    LLMClient abstracts all the fixed arguments to process_prompts_async, so you can create it
-    once and use it for more stuff without having to configure all the arguments.
-    Handles models, sampling params for each model, model weights, rate limits, etc.
+    Internal LLMClient implementation using Pydantic.
+    Keeps all validation, serialization, and existing functionality.
     """
 
-    model_names: list[str] = ["gpt-4.1-mini"]
-
-    def __init__(self, model_name: str | list[str] | None = None, **kwargs):
-        if model_name is not None:
-            kwargs["model_names"] = model_name
-        super().__init__(**kwargs)
-
+    model_names: str | list[str] = ["gpt-4.1-mini"]
     max_requests_per_minute: int = 1_000
     max_tokens_per_minute: int = 100_000
     max_concurrent_requests: int = 225
@@ -59,6 +50,36 @@ class LLMClient(BaseModel):
     top_logprobs: int | None = None
     force_local_mcp: bool = False
 
+    # Progress configuration
+    progress: Literal["rich", "tqdm", "manual"] = "rich"
+
+    # Postprocessing - run on every APIResponse
+    postprocess: Callable[[APIResponse], APIResponse] | None = None
+
+    # Internal state for async task handling
+    _next_task_id: int = PrivateAttr(default=0)
+    _tasks: dict[int, asyncio.Task] = PrivateAttr(default_factory=dict)
+    _results: dict[int, APIResponse] = PrivateAttr(default_factory=dict)
+    _tracker: StatusTracker | None = PrivateAttr(default=None)
+    _capacity_lock: asyncio.Lock = PrivateAttr(default_factory=asyncio.Lock)
+
+    # Progress management for queueing API
+    def open(self, total: int | None = None, show_progress: bool = True):
+        self._tracker = StatusTracker(
+            max_requests_per_minute=self.max_requests_per_minute,
+            max_tokens_per_minute=self.max_tokens_per_minute,
+            max_concurrent_requests=self.max_concurrent_requests,
+            progress_style=self.progress,
+            use_progress_bar=show_progress,
+        )
+        self._tracker.init_progress_bar(total)
+        return self
+
+    def close(self):
+        if self._tracker:
+            self._tracker.log_final_status()
+            self._tracker = None
+
     # NEW! Builder methods
     def with_model(self, model: str):
         self.model_names = [model]
@@ -81,15 +102,23 @@ class LLMClient(BaseModel):
         if max_concurrent_requests:
             self.max_concurrent_requests = max_concurrent_requests
 
+    def _get_tracker(self) -> StatusTracker:
+        if self._tracker is None:
+            self.open()
+            assert self._tracker, "should have tracker now"
+        return self._tracker
+
     @property
     def models(self):
         return self.model_names  # why? idk
 
     @model_validator(mode="before")
     @classmethod
-    def fix_lists(cls, data) -> "LLMClient":
+    def fix_lists(cls, data) -> "_LLMClient":
         if isinstance(data.get("model_names"), str):
             data["model_names"] = [data["model_names"]]
+        if not isinstance(data.get("sampling_params", []), list):
+            data["sampling_params"] = [data["sampling_params"]]
         if "sampling_params" not in data or len(data.get("sampling_params", [])) == 0:
             data["sampling_params"] = [
                 SamplingParams(
@@ -162,14 +191,6 @@ class LLMClient(BaseModel):
         config_dict = yaml.safe_load(open(file_path))
         return cls.from_dict(config_dict)
 
-    @classmethod
-    def basic(cls, model: str | list[str], **kwargs):
-        """
-        Doesn't do anything differently now, kept for backwards compat.
-        """
-        kwargs["model_names"] = model
-        return cls(**kwargs)
-
     def _select_model(self):
         assert isinstance(self.model_weights, list)
         model_idx = np.random.choice(range(len(self.models)), p=self.model_weights)
@@ -192,16 +213,23 @@ class LLMClient(BaseModel):
         chosen_sp = self.sampling_params[self.models.index(chosen_model)]
         return chosen_model, chosen_sp
 
-    async def _wait_for_capacity(self, num_tokens: int, tracker: StatusTracker):
+    async def _wait_for_capacity(
+        self, num_tokens: int, tracker: StatusTracker, *, retry: bool = False
+    ):
         while True:
-            if tracker.check_capacity(num_tokens):
-                tracker.set_limiting_factor(None)
-                return
+            # Enforce cooldown first, regardless of current capacity.
+            cooldown = tracker.seconds_to_pause
+            if cooldown > 0:
+                print(f"Pausing for {cooldown} seconds to cool down.")
+                await asyncio.sleep(cooldown)
+                continue
 
-            if tracker.seconds_to_pause > 0:
-                await asyncio.sleep(tracker.seconds_to_pause)
-            else:
-                await asyncio.sleep(random.random())
+            async with self._capacity_lock:
+                if tracker.check_capacity(num_tokens, retry=retry):
+                    tracker.set_limiting_factor(None)
+                    return
+            # Idle wait before next capacity check. Aim for ~RPM spacing.
+            await asyncio.sleep(max(60.0 / self.max_requests_per_minute, 0.01))
 
     async def _execute_request(self, context: RequestContext) -> APIResponse:
         """Create and send a single API request using the provided context."""
@@ -214,14 +242,20 @@ class LLMClient(BaseModel):
         self, context: RequestContext, retry_queue: asyncio.Queue | None = None
     ) -> APIResponse:
         """Handle caching and single HTTP call for a request. Failed requests go to retry queue."""
+
         # Check cache first
+        def _maybe_postprocess(response: APIResponse):
+            if self.postprocess:
+                return self.postprocess(response)
+            return response
+
         if self.cache:
             cached = self.cache.get(context.prompt)
             if cached:
                 cached.local_cache_hit = True
                 if context.status_tracker:
                     context.status_tracker.task_succeeded(context.task_id)
-                return cached
+                return _maybe_postprocess(cached)
 
         # Execute single request
         assert context.status_tracker
@@ -236,7 +270,7 @@ class LLMClient(BaseModel):
                 self.cache.put(context.prompt, response)
             # Call callback if provided
             context.maybe_callback(response, context.status_tracker)
-            return response
+            return _maybe_postprocess(response)
 
         # Handle error response - add to retry queue if available
         if retry_queue and context.attempts_left > 1:
@@ -255,7 +289,7 @@ class LLMClient(BaseModel):
 
             # Print error message for debugging
             error_msg = (
-                f"Error task {context.task_id}. Model: {response.model_internal}"
+                f"😔 Error task {context.task_id}. Model: {response.model_internal}"
             )
             if response.status_code:
                 error_msg += f" Code: {response.status_code},"
@@ -264,7 +298,7 @@ class LLMClient(BaseModel):
 
             # Add to retry queue for later processing
             await retry_queue.put(retry_context)
-            return response  # Return the error response for now
+            return _maybe_postprocess(response)  # Return the error response for now
 
         # No retries left or no retry queue - final failure
         context.status_tracker.task_failed(context.task_id)
@@ -277,7 +311,7 @@ class LLMClient(BaseModel):
         error_msg += f" Message: {response.error_message}. Giving up."
         print(error_msg)
 
-        return response
+        return _maybe_postprocess(response)
 
     @overload
     async def process_prompts_async(
@@ -317,38 +351,35 @@ class LLMClient(BaseModel):
         prompts = prompts_to_conversations(prompts)
         ids = list(range(len(prompts)))
         results: list[APIResponse | None] = [None for _ in range(len(prompts))]
-
-        # Create StatusTracker
-        tracker = StatusTracker(
-            max_requests_per_minute=self.max_requests_per_minute,
-            max_tokens_per_minute=self.max_tokens_per_minute,
-            max_concurrent_requests=self.max_concurrent_requests,
-            use_progress_bar=show_progress,
-            progress_bar_total=len(prompts),
-            progress_bar_disable=not show_progress,
-            use_rich=show_progress,
-        )
-
-        tracker.init_progress_bar()
+        # Use existing tracker if client has been opened; otherwise open/close automatically
+        tracker: StatusTracker
+        tracker_preopened = self._tracker is not None
+        if tracker_preopened:
+            tracker = self._tracker  # type: ignore[assignment]
+            tracker.add_to_total(len(prompts))
+        else:
+            self.open(total=len(prompts), show_progress=show_progress)
+            tracker = self._tracker  # type: ignore[assignment]
+        assert tracker is not None
 
         # Create retry queue for failed requests
         retry_queue: asyncio.Queue[RequestContext] = asyncio.Queue()
 
-        # Calculate sleep time for rate limiting
+        # Calculate sleep time for rate limiting (legacy; gating happens in _wait_for_capacity)
         seconds_to_sleep_each_loop = (60.0 * 0.9) / tracker.max_requests_per_minute
 
         # Main dispatch loop - using original pattern but with all prompts
         next_context = None  # Persist across iterations like original
+        next_is_retry = False  # Track whether next_context is a retry
         prompts_not_finished = True
         prompts_iter = iter(zip(ids, prompts))
 
         while True:
             # Get next context (retry or new) - only if we don't already have one waiting
-            retry_request = False
             if next_context is None:
                 if not retry_queue.empty():
                     next_context = retry_queue.get_nowait()
-                    retry_request = True
+                    next_is_retry = True
                     print(f"Retrying request {next_context.task_id}.")
                 elif prompts_not_finished:
                     try:
@@ -369,58 +400,58 @@ class LLMClient(BaseModel):
                             extra_headers=self.extra_headers,
                             force_local_mcp=self.force_local_mcp,
                         )
+
+                        next_is_retry = False
                     except StopIteration:
                         prompts_not_finished = False
 
-            # Update capacity - original logic
-            tracker.update_capacity()
-
-            # Dispatch if capacity available - original logic
+            # Dispatch using shared capacity gate (consistent with start_nowait)
             if next_context:
-                if tracker.check_capacity(next_context.num_tokens, retry=retry_request):
-                    tracker.set_limiting_factor(None)
+                # Wait here until we have capacity to launch this context
+                await self._wait_for_capacity(
+                    next_context.num_tokens, tracker, retry=next_is_retry
+                )
 
-                    # Launch simplified request processing
-                    async def process_and_store(ctx: RequestContext):
-                        try:
-                            response = await self.process_single_request(
-                                ctx, retry_queue
-                            )
-                            results[ctx.task_id] = response
-                        except Exception as e:
-                            # Create an error response for validation errors and other exceptions
-                            error_response = APIResponse(
-                                id=ctx.task_id,
-                                model_internal=ctx.model_name,
-                                prompt=ctx.prompt,
-                                sampling_params=ctx.sampling_params,
-                                status_code=None,
-                                is_error=True,
-                                error_message=str(e),
-                            )
-                            results[ctx.task_id] = error_response
-                            # Mark task as completed so the main loop can finish
-                            if ctx.status_tracker:
-                                ctx.status_tracker.task_failed(ctx.task_id)
+                # Launch simplified request processing
+                async def process_and_store(ctx: RequestContext):
+                    try:
+                        response = await self.process_single_request(ctx, retry_queue)
+                        results[ctx.task_id] = response
+                    except Exception as e:
+                        # Create an error response for validation errors and other exceptions
+                        error_response = APIResponse(
+                            id=ctx.task_id,
+                            model_internal=ctx.model_name,
+                            prompt=ctx.prompt,
+                            sampling_params=ctx.sampling_params,
+                            status_code=None,
+                            is_error=True,
+                            error_message=str(e),
+                        )
+                        results[ctx.task_id] = error_response
+                        # Mark task as completed so the main loop can finish
+                        if ctx.status_tracker:
+                            ctx.status_tracker.task_failed(ctx.task_id)
 
-                    asyncio.create_task(process_and_store(next_context))
-                    next_context = None  # Reset after successful dispatch
+                asyncio.create_task(process_and_store(next_context))
+                next_context = None  # Reset after successful dispatch
+                next_is_retry = False
 
             # Update progress - original logic
             tracker.update_pbar()
 
-            # Check completion - original logic
-            if (
-                tracker.num_tasks_in_progress == 0
-                and not prompts_not_finished
-                and retry_queue.empty()
-            ):
+            # Check completion: consider final outcomes, not in-progress count
+            # This avoids rare hangs if in-progress is miscounted (e.g., double-increment).
+            if (tracker.num_tasks_succeeded + tracker.num_tasks_failed) >= len(
+                prompts
+            ) and retry_queue.empty():
                 break
 
-            # Sleep - original logic
-            await asyncio.sleep(seconds_to_sleep_each_loop + tracker.seconds_to_pause)
+            # Yield briefly to allow in-flight tasks to progress
+            await asyncio.sleep(min(0.01, seconds_to_sleep_each_loop))
 
-        tracker.log_final_status()
+        if not tracker_preopened:
+            self.close()
 
         if return_completions_only:
             return [r.completion if r is not None else None for r in results]
@@ -435,6 +466,7 @@ class LLMClient(BaseModel):
         show_progress=True,
         tools: list[Tool | dict | MCPServer] | None = None,
         cache: CachePattern | None = None,
+        use_responses_api: bool = False,
     ):
         return asyncio.run(
             self.process_prompts_async(
@@ -443,9 +475,119 @@ class LLMClient(BaseModel):
                 show_progress=show_progress,
                 tools=tools,
                 cache=cache,
+                use_responses_api=use_responses_api,
             )
         )
 
+    async def _run_context(self, context: RequestContext) -> APIResponse:
+        tracker = self._get_tracker()
+        retry = False
+        retry_queue: asyncio.Queue[RequestContext] = asyncio.Queue()
+        current = context
+        while True:
+            await self._wait_for_capacity(current.num_tokens, tracker, retry=retry)
+            response = await self.process_single_request(current, retry_queue)
+            if not response.is_error or retry_queue.empty():
+                self._results[context.task_id] = response
+                return response
+            current = await retry_queue.get()
+            retry = True
+
+    def start_nowait(
+        self,
+        prompt: str | Conversation,
+        *,
+        tools: list[Tool | dict | MCPServer] | None = None,
+        cache: CachePattern | None = None,
+        use_responses_api: bool = False,
+    ) -> int:
+        tracker = self._get_tracker()
+        task_id = self._next_task_id
+        self._next_task_id += 1
+        model, sampling_params = self._select_model()
+        if isinstance(prompt, str):
+            prompt = Conversation.user(prompt)
+        context = RequestContext(
+            task_id=task_id,
+            model_name=model,
+            prompt=prompt,
+            sampling_params=sampling_params,
+            attempts_left=self.max_attempts,
+            request_timeout=self.request_timeout,
+            status_tracker=tracker,
+            tools=tools,
+            cache=cache,
+            use_responses_api=use_responses_api,
+            extra_headers=self.extra_headers,
+            force_local_mcp=self.force_local_mcp,
+        )
+        task = asyncio.create_task(self._run_context(context))
+        self._tasks[task_id] = task
+        tracker.add_to_total(1)
+        return task_id
+
+    async def start(
+        self,
+        prompt: str | Conversation,
+        *,
+        tools: list[Tool | dict | MCPServer] | None = None,
+        cache: CachePattern | None = None,
+        use_responses_api: bool = False,
+    ) -> APIResponse | None:
+        task_id = self.start_nowait(
+            prompt, tools=tools, cache=cache, use_responses_api=use_responses_api
+        )
+        return await self.wait_for(task_id)
+
+    async def wait_for(self, task_id: int) -> APIResponse | None:
+        task = self._tasks.get(task_id)
+        if task:
+            return await task
+        return self._results.get(task_id)
+
+    async def wait_for_all(
+        self, task_ids: Sequence[int] | None = None
+    ) -> list[APIResponse | None]:
+        if task_ids is None:
+            task_ids = list(self._tasks.keys())
+        return [await self.wait_for(tid) for tid in task_ids]
+
+    async def as_completed(
+        self, task_ids: Sequence[int] | None = None
+    ) -> AsyncGenerator[tuple[int, APIResponse | None], None]:
+        """Yield ``(task_id, result)`` pairs as tasks complete.
+
+        Args:
+            task_ids: Optional sequence of task IDs to wait on. If ``None``,
+                all queued tasks are watched.
+
+        Yields:
+            Tuples of task ID and ``APIResponse`` as each task finishes.
+        """
+
+        if task_ids is None:
+            tasks_map: dict[asyncio.Task, int] = {
+                task: tid for tid, task in self._tasks.items()
+            }
+        else:
+            tasks_map = {
+                self._tasks[tid]: tid for tid in task_ids if tid in self._tasks
+            }
+
+        # Yield any tasks that have already completed
+        for task in list(tasks_map.keys()):
+            if task.done():
+                tid = tasks_map.pop(task)
+                yield tid, self._results.get(tid, await task)
+
+        while tasks_map:
+            done, _ = await asyncio.wait(
+                set(tasks_map.keys()), return_when=asyncio.FIRST_COMPLETED
+            )
+            for task in done:
+                tid = tasks_map.pop(task)
+                yield tid, self._results.get(tid, await task)
+
     async def stream(
         self,
         prompt: str | Conversation,
@@ -461,6 +603,8 @@ class LLMClient(BaseModel):
                 print(item, end="", flush=True)
             else:
                 # final item
+                if self.postprocess:
+                    return self.postprocess(item)
                 return item
 
     async def run_agent_loop(
@@ -505,7 +649,7 @@ class LLMClient(BaseModel):
             if last_response is None or last_response.content is None:
                 break
 
-            conversation.add(last_response.content)
+            conversation = conversation.with_message(last_response.content)
 
             tool_calls = last_response.content.tool_calls
             if not tool_calls:
@@ -604,59 +748,117 @@ class LLMClient(BaseModel):
         )
 
 
-# def api_prompts_dry_run(
-#     ids: np.ndarray | list[int],
-#     prompts: list[Conversation],
-#     models: str | list[str],
-#     model_weights: list[float],
-#     sampling_params: list[SamplingParams],
-#     max_tokens_per_minute: int = 500_000,
-#     max_requests_per_minute: int = 1_000,
-# ):
-#     """
-#     Count tokens and estimate costs for a batch of prompts.
-#     """
-#     results = []
-#     for i, prompt in zip(ids, prompts):
-#         # choose a model
-#         model_idx = np.random.choice(range(len(models)), p=model_weights)
-#         model = models[model_idx]
+# factory function -- allows positional model names,
+# keeps pydantic validation, without sacrificing IDE support
+@overload
+def LLMClient(
+    model_names: str,
+    *,
+    max_requests_per_minute: int = 1_000,
+    max_tokens_per_minute: int = 100_000,
+    max_concurrent_requests: int = 225,
+    sampling_params: list[SamplingParams] | None = None,
+    model_weights: list[float] | Literal["uniform", "dynamic"] = "uniform",
+    max_attempts: int = 5,
+    request_timeout: int = 30,
+    cache: Any = None,
+    extra_headers: dict[str, str] | None = None,
+    temperature: float = 0.75,
+    top_p: float = 1.0,
+    json_mode: bool = False,
+    max_new_tokens: int = 512,
+    reasoning_effort: Literal["low", "medium", "high", None] = None,
+    logprobs: bool = False,
+    top_logprobs: int | None = None,
+    force_local_mcp: bool = False,
+    progress: Literal["rich", "tqdm", "manual"] = "rich",
+    postprocess: Callable[[APIResponse], APIResponse] | None = None,
+) -> _LLMClient: ...
 
-#         # dry run
-#         input_tokens, output_tokens, min_cost, max_cost = prompt.dry_run(
-#             model, sampling_params[model_idx].max_new_tokens
-#         )
-#         results.append(
-#             {
-#                 "id": i,
-#                 "input_tokens": input_tokens,
-#                 "output_tokens": output_tokens,
-#                 "min_cost": min_cost,
-#                 "max_cost": max_cost,
-#             }
-#         )
 
-#     combined_results: dict[str, Any] = {
-#         "total_input_tokens": sum([r["input_tokens"] for r in results]),
-#         "total_output_tokens": sum([r["output_tokens"] for r in results]),
-#         "total_min_cost": sum([r["min_cost"] for r in results]),
-#         "total_max_cost": sum([r["max_cost"] for r in results]),
-#     }
-#     minimum_time_tpm = combined_results["total_input_tokens"] / max_tokens_per_minute
-#     maximum_time_tpm = (
-#         combined_results["total_input_tokens"] + combined_results["total_output_tokens"]
-#     ) / max_tokens_per_minute
-#     minimum_time_rpm = len(prompts) / max_requests_per_minute
+@overload
+def LLMClient(
+    model_names: list[str],
+    *,
+    max_requests_per_minute: int = 1_000,
+    max_tokens_per_minute: int = 100_000,
+    max_concurrent_requests: int = 225,
+    sampling_params: list[SamplingParams] | None = None,
+    model_weights: list[float] | Literal["uniform", "dynamic"] = "uniform",
+    max_attempts: int = 5,
+    request_timeout: int = 30,
+    cache: Any = None,
+    extra_headers: dict[str, str] | None = None,
+    temperature: float = 0.75,
+    top_p: float = 1.0,
+    json_mode: bool = False,
+    max_new_tokens: int = 512,
+    reasoning_effort: Literal["low", "medium", "high", None] = None,
+    logprobs: bool = False,
+    top_logprobs: int | None = None,
+    force_local_mcp: bool = False,
+    progress: Literal["rich", "tqdm", "manual"] = "rich",
+    postprocess: Callable[[APIResponse], APIResponse] | None = None,
+) -> _LLMClient: ...
 
-#     combined_results["minimum_time"] = max(minimum_time_tpm, minimum_time_rpm)
-#     combined_results["maximum_time"] = max(maximum_time_tpm, minimum_time_rpm)
-#     limiting_factor = None
-#     if minimum_time_rpm > maximum_time_tpm:
-#         limiting_factor = "requests"
-#     elif minimum_time_rpm < minimum_time_tpm:
-#         limiting_factor = "tokens"
-#     else:
-#         limiting_factor = "depends"
-#     combined_results["limiting_factor"] = limiting_factor
 
-#     return combined_results
+def LLMClient(
+    model_names: str | list[str] = "gpt-4.1-mini",
+    *,
+    max_requests_per_minute: int = 1_000,
+    max_tokens_per_minute: int = 100_000,
+    max_concurrent_requests: int = 225,
+    sampling_params: list[SamplingParams] | None = None,
+    model_weights: list[float] | Literal["uniform", "dynamic"] = "uniform",
+    max_attempts: int = 5,
+    request_timeout: int = 30,
+    cache: Any = None,
+    extra_headers: dict[str, str] | None = None,
+    temperature: float = 0.75,
+    top_p: float = 1.0,
+    json_mode: bool = False,
+    max_new_tokens: int = 512,
+    reasoning_effort: Literal["low", "medium", "high", None] = None,
+    logprobs: bool = False,
+    top_logprobs: int | None = None,
+    force_local_mcp: bool = False,
+    progress: Literal["rich", "tqdm", "manual"] = "rich",
+    postprocess: Callable[[APIResponse], APIResponse] | None = None,
+) -> _LLMClient:
+    """
+    Create an LLMClient with model_names as a positional argument.
+
+    Args:
+        model_names: Model name(s) to use - can be a single string or list of strings
+        **kwargs: All other LLMClient configuration options (keyword-only)
+
+    Returns:
+        Configured LLMClient instance
+    """
+    # Handle default for mutable argument
+    if sampling_params is None:
+        sampling_params = []
+
+    # Simply pass everything to the Pydantic constructor
+    return _LLMClient(
+        model_names=model_names,
+        max_requests_per_minute=max_requests_per_minute,
+        max_tokens_per_minute=max_tokens_per_minute,
+        max_concurrent_requests=max_concurrent_requests,
+        sampling_params=sampling_params,
+        model_weights=model_weights,
+        max_attempts=max_attempts,
+        request_timeout=request_timeout,
+        cache=cache,
+        extra_headers=extra_headers,
+        temperature=temperature,
+        top_p=top_p,
+        json_mode=json_mode,
+        max_new_tokens=max_new_tokens,
+        reasoning_effort=reasoning_effort,
+        logprobs=logprobs,
+        top_logprobs=top_logprobs,
+        force_local_mcp=force_local_mcp,
+        progress=progress,
+        postprocess=postprocess,
+    )
